#define PickType() ( (++color)&3 )

const int kErrorCheckGL = 0 ;

#include "shared/core/first.h"
#include "ext/GL/glut.h"
#include "shared/graphics/gl/gl_util.h"
#include "shared/core/console.h"
#include "shared/graphics/gl/console_gl.h"
#include "shared/input/mouse.h"
#include "shared/input/keyboard.h"
#include "shared/input/joystick.h"
#include "shared/core/adjuster.h"

#include "shared/physics/2d/gl_physics2d_draw.h"
#include "shared/physics/2d/rigid/2d_rigid_sim.h"
#include "shared/physics/2d/rigid/2d_rigid_drag_force.h"
#include "shared/physics/2d/rigid/2d_rigid_gravity_force.h"
#include "shared/physics/2d/rigid/2d_rigid_spring_force.h"
#include "shared/physics/2d/rigid/2d_rigid_pt2pt_constraint.h"

using namespace ajb;

//---
#include "chaim.h"
#include "SimHelp.h"

class glutApp
{
public:
	glutApp()
	{
		isMouseDown = false ;
	}

	float mousex, mousey ;
	float lastMouseDownX, lastMouseDownY ;
	bool wasClickInPlace, isMouseDown  ;

} theGLUTApp ;

fRect GetBounds ( void ) { return fRect(0,0,kWidth,kHeight) ; }

bool	GetOButton	   ( int j )
{
	return JOYSTICK.button_went_down( Joystick::BUTTON_RIGHT_R, j ) ;
}

bool	GetXButton	   ( int j )
{
	return JOYSTICK.button_went_down( Joystick::BUTTON_RIGHT_D, j ) ||
		   JOYSTICK.button_went_down( Joystick::BUTTON_RIGHT_U, j ) ||
		   JOYSTICK.button_went_down( Joystick::BUTTON_RIGHT_L, j ) ||
		   JOYSTICK.button_went_down( Joystick::BUTTON_RIGHT_R, j ) ;
}

fPoint2 GetLeftStick ( int j )
{
	fPoint2 l = v2f( JOYSTICK.left_stick(j) ) ;
	fPoint2 r = v2f( JOYSTICK.right_stick(j) ) ;

	fPoint2 p = l ;
	p.Add(r) ;

	p.y *= -1 ;

	fRect g(-1,-1,1,1) ;
	
	p = g.PinPointInside(p) ;

	return p ;
}

bool	GetSelectButton( int j )
{
	return JOYSTICK.button_went_down( Joystick::BUTTON_SELECT, j ) ;
}

bool GetStartButton ( int j )
{
	return JOYSTICK.button_went_down( Joystick::BUTTON_START, j ) ;
}

bool GetIsMouseDown ( void )
{
	return theGLUTApp.isMouseDown ;
}

fPoint2 GetMouseDownLoc ( void )
{
	fRect q = fRect(0,0,kPixelWidth,kPixelHeight) ;

	fPoint2 p( theGLUTApp.lastMouseDownX, theGLUTApp.lastMouseDownY ) ;

	p = q.PointToFraction(p) ;

	return GetBounds().FractionToPoint(p) ;
}

fPoint2 GetMouseLoc ( void )
{
	fRect q = fRect(0,0,kPixelWidth,kPixelHeight) ;

	fPoint2 p( theGLUTApp.mousex, theGLUTApp.mousey ) ;

	p = q.PointToFraction(p) ;

	return GetBounds().FractionToPoint(p) ;
}

// ---

extern void launchEditor(char *s);
extern void loadPhysicsFile(char *str, float x, float y, float scale, float density_scale,
						    ajb::Physics::D2::Rigid::Rigid_Body   ***rb_list, int *num_bodies,
						    ajb::Physics::D2::Rigid::Binary_Force *** f_list, int *num_forces,
						    ajb::Physics::D2::Rigid::Rigid_Body *world);

namespace ajb {

using namespace Physics::D2;
using namespace Physics::D2::Rigid;

int G_win_width = 0;
int G_win_height = 0;
Adjuster_Set adjusters;
GL::Console_GL* console = 0;
float screen_width = 100.0f;
Vec2 camera_pos = Vec2(0.0f,65.0f);   // im offseting the original camera pos

Sim* sim = 0;
Drag_Force* drag = 0;
Gravity_Force* gravity = 0;
Spring_Force* spring = 0;
float obj_size = 4.0f;
float grav_value = -10.0f;

int color = 0;        // what color will the next object be
int frame_count = 0;
int last_create = 0;  // last frame we created an object
int points = 0;
int drop_counter = 0;
int obj_count = 0;

int paused = 0;
int game_running = 0;

Rigid_Body *world_left, *world_right, *world_base;

float color_map[9][3]=
{{1,0,0},{0,1,0},{0,0,1},{1,1,0},{1,0,1},{0,1,1},{0.2,0.8,0.2},{0.2,0.2,0.8},{0.8,0.2,0.2}};

#define MAX_ATTACHED 16

class cGameObj {
public:
   Rigid_Body    *rb;
   unsigned char  flags;
   char           type;    // 0 - 7 are world objs
   char           shape;
   char           attach_count;
   Angle_Spring_Force *ang[MAX_ATTACHED];
   Pt2Pt_Constraint   *cnt[MAX_ATTACHED];
   cGameObj      *attached_list[MAX_ATTACHED];  // are we attached to anyone?
};

#define kGameObjInUse  0x01
#define kGameObjScored 0x02
#define MAX_GAME_OBJ   1024
cGameObj gGameObjs[MAX_GAME_OBJ];  // for now, default to 0 rb pointer means "free"

cGameObj *alloc_GameObj(void)
{
   for (int i=0; i<MAX_GAME_OBJ; i++)
      if ((gGameObjs[i].flags&kGameObjInUse)==0)
      {
         obj_count++;
         gGameObjs[i].flags=kGameObjInUse;
         gGameObjs[i].attach_count=0;
         for (int j=0; j>MAX_ATTACHED; j++)
         {
            gGameObjs[i].cnt[j]=0;
            gGameObjs[i].ang[j]=0;
            gGameObjs[i].attached_list[i]=0;
         }
         return &gGameObjs[i];
      }
   return 0;
}

void free_GameObj(cGameObj *pObj)
{
   if ((pObj->flags&kGameObjInUse)==0)
   {
      // @TODO: assert
   }
   else if (pObj->rb)
   {
      for (int i=0; i<MAX_ATTACHED; i++)
      {  // dont recurse attached list, but do kill constraints
         if (pObj->ang[i])
         {
            sim->remove(*(pObj->ang[i]));
            delete pObj->ang[i];
         }
         if (pObj->cnt[i])
         {
            sim->remove(*(pObj->cnt[i]));
            delete pObj->cnt[i];
         }
      }
      sim->remove(*(pObj->rb));
      delete pObj->rb;  // assume we allocated it
      pObj->rb=0;
      pObj->flags=0;
      obj_count--;
   }
}

// this works only if you promise to kill everything connected to anyone in here
// i.e. need a better/real/robust one at some point
void del_ObjGroup(cGameObj *pObj)
{
   for (int i=0; i<MAX_ATTACHED; i++)
   {
      if (pObj->attached_list[i]!=0)
         free_GameObj(pObj->attached_list[i]);
   }
   free_GameObj(pObj);
}

float frand(void)
{
   return ((float)(rand()&0x7fff))/(float)0x7fff;
}

void change_gravity(float new_grav)
{
   if (gravity)
   {
      sim->remove(*gravity);
      delete gravity;
   }
   gravity = new Gravity_Force(Vec2(0.0f, new_grav));
   sim->add(*gravity);
}

void set_drop_counter(void)
{
   drop_counter=512-(frame_count/512);
   if (drop_counter<32) drop_counter=32;
}

int game_over(void)
{
   return 0;
}

void _bg_setup(void)
{
   // all world bodies have Zero inverse mass and inverse inertia
   world_left  = new Rigid_Body(make_Body_Box(Vec2(-2.5f, 0.0f), Vec2(2.5f, 120.0f)),
                                 Vec2(-77.5f,0.0f),0.0f,0.0f,0.0f);
   world_base  = new Rigid_Body(make_Body_Box(Vec2(-80.0f, -2.5f), Vec2(80.0f, +2.5f)),
                                 Vec2::zero(),0.0f,0.0f,0.0f);
   world_right = new Rigid_Body(make_Body_Box(Vec2(-2.5f, 0.0f), Vec2(2.5f, 120.0f)),
                                 Vec2(+77.5f,0.0f),0.0f,0.0f,0.0f);
   // rewrite this (right) to reuse the body_box from left?

   sim->add(*world_left);
   sim->add(*world_base);
   sim->add(*world_right);
}

// give some initial rotation, too...
cGameObj *add_obj(const Vec2& src_pt, const int type, const float size)
{
   // make_Body_Sphere(size);
   cGameObj *pObj = alloc_GameObj();

   if (pObj)
   {
      float w=(frand()-0.5f)*3;

      pObj->rb = new Rigid_Body(make_Body_Box(Vec2(-size,-size),Vec2(size,size)), src_pt, 0.0f, 1.0f, 1.0f, Vec2::zero(), w);
      pObj->type = type;
      sim->add(*(pObj->rb));
      pObj->rb->m_pAppData=pObj;
   }
   else
   {
      // @TODO: assert
   }
   return pObj;
}

void load_file_test(void)
{
   // Load database from file
   {
		int num_rb=0, num_force=0;
		Rigid_Body **rb_list;
		Binary_Force **force_list;
		loadPhysicsFile("bodies.physdb", 0,22, 1.0/16, 1.0/16, &rb_list, &num_rb, &force_list, &num_force, world_base);
		for (int i=0; i < num_rb; ++i) {
			sim->add(*rb_list[i]);
		}
		for (int i=0; i < num_force; ++i) {
			sim->add(*force_list[i]);
		}
   }
}

void game_init()
{
	sim = new Sim;

	drag = new Drag_Force(.05f, .05f);	//a little drag helps with stability
	sim->add(*drag);
    change_gravity(grav_value);

    _bg_setup();
//    add_obj(Vec2(0.0f,10.0f),8,1);  // initial box of fun, tm

	InitChaim() ;
}

void PositionCamera ( void )
{
	GL::gl_error_check();
	glMatrixMode( GL_PROJECTION ) ;
	glLoadIdentity() ;

	float w = kWidth, h = kHeight ;
	float cameraZoom = 1.0 ;
	float camerax = w/2, cameray = h/2 ;

	glOrtho( 0, w, h, 0, -1, 1 ) ;
	glTranslatef( w/2, h/2, 0 ) ;

	GL::gl_error_check();
	glScalef( cameraZoom, cameraZoom, 1.0 ) ;

	glTranslatef( -w/2, -h/2, 0 ) ;

	glMatrixMode( GL_MODELVIEW ) ;
	glLoadIdentity() ;

	glTranslatef( camerax, cameray, 0 ) ;

	glTranslatef( -w/2, -h/2, 0 ) ;
	GL::gl_error_check();
}

void game_begin_frame()
{
	console->clear();
//	AJB_CONSOLE(Heap::stats_dump);	//probably want to leave these commented out until things get too slow
//	AJB_CONSOLE(Hier_Profile_dump);
	Hier_Profiler_clear();
	Heap::reset_stats();

	PROFILER(begin_frame);
	if (kErrorCheckGL) GL::gl_error_check();
	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	PositionCamera() ;

	JOYSTICK.sample();
	if (KEYBOARD.is_down(27)) //escape
		exit(1);

// 	adjusters.tick(
// 		KEYBOARD.went_down(KEYBOARD_KEY_UP   ), KEYBOARD.went_down(KEYBOARD_KEY_DOWN),	//up, down
// 		KEYBOARD.went_down(KEYBOARD_KEY_LEFT ), KEYBOARD.is_down  (KEYBOARD_KEY_LEFT),	//less
// 		KEYBOARD.went_down(KEYBOARD_KEY_RIGHT), KEYBOARD.is_down  (KEYBOARD_KEY_RIGHT)	//more
// 	);
// 	adjusters.dump(*AJB_CONSOLE_PTR);
}

void game_tick()
{
	PROFILER(tick);

   frame_count++;

	screen_width += 1.0f*(KEYBOARD.is_down('z') - KEYBOARD.is_down('a'));
	if (screen_width < .1f) screen_width = .1f;
	camera_pos += Vec2(
		0.02f*screen_width*(KEYBOARD.is_down('f') - KEYBOARD.is_down('s')),
		0.02f*screen_width*(KEYBOARD.is_down('e') - KEYBOARD.is_down('d'))
	);

	const Vec2 mouse_screen_space = MOUSE.get_pos();
	const Vec2 mouse_world_space(
		camera_pos.x() + (mouse_screen_space.x()-0.5f*G_win_width )*(2* screen_width/G_win_width), 
		camera_pos.y() + (mouse_screen_space.y()-0.5f*G_win_height)*(2*-screen_width/G_win_width)
 	);

//	AJB_CONSOLE("mouse screen space " << mouse_screen_space.x() << " " << mouse_screen_space.y() << endl);
//	AJB_CONSOLE("mouse world  space " << mouse_world_space. x() << " " << mouse_world_space. y() << endl);
//	AJB_CONSOLE("camera_pos " << camera_pos.x() << " " << camera_pos.y() << endl);
	
	if (MOUSE.left_went_down()) {			//attach
		Vec2 body_space_position;
		Rigid_Body* rb = sim->pick(body_space_position, mouse_world_space);
		if (rb) {
			spring = new Spring_Force(
				Handle(*world_base, mouse_world_space),
				Handle(*rb, body_space_position), 
				4.0f,		//Kspring
				2.0f,		//Kdamp
				0.0f		//rest length
			);
			sim->add(*spring);
		}
	}
	if (spring && !MOUSE.left_is_down()) {	//release
		sim->remove(*spring);
		delete spring;
		spring = 0;
	}

	if (spring)								//update
       spring->m_a = Handle(*world_base, mouse_world_space);

	//if (MOUSE.right_went_down()||(KEYBOARD.is_down('n')&&(frame_count>last_create+15)))
   if ( frame_count%30 == 1 )
	{
      add_obj(mouse_world_space, PickType(),obj_size);
      last_create=frame_count;
   }

   if (KEYBOARD.went_down('p'))
      paused ^= 1;
   if (KEYBOARD.went_down('0'))
      game_running ^= 1;

   if (KEYBOARD.went_down('l'))
   {
      static int loaded=0;
      if (!loaded)
      {
         AJB_CONSOLE("Loading " << endl);
         load_file_test();
         loaded=1;
      }
   }


   if (!paused)
   {
      if (game_running)
         if (drop_counter<=0)
         {
            Vec2  new_pos((frand()-0.5)*70,110.f);
            float new_size=2+(frand()*3);
            add_obj(new_pos, PickType(), new_size);
            last_create=frame_count;
            set_drop_counter();
         }
         else
            drop_counter--;

      const int num_sim_steps = 2;
      for (int i=0; i < num_sim_steps; ++i)
      {
         sim->step(1.0f/60.0f*(1.0f/num_sim_steps));
         
         // lets go have fun parsing our contact list, sim->m_contacts
         for (Contact_List::const_iterator cit=sim->m_contacts.begin(); cit != sim->m_contacts.end(); ++cit) 
         {
            //         const Contact& c = *cit;
            cGameObj *pObjA = (cGameObj *) (((Rigid_Body *)(cit->m_Prb_a))->m_pAppData);
            cGameObj *pObjB = (cGameObj *) (((Rigid_Body *)(cit->m_Prb_b))->m_pAppData);
            if (pObjA!=0 && pObjB!=0)  // have to do this dumb 2nd check since we can delete while we loop
               if ((pObjA->flags&kGameObjInUse)&&(pObjB->flags&kGameObjInUse))
                  if (pObjA->type==pObjB->type)  // same type colliding - maybe points
                  { 
                     const float PiOver2=3.1415926535f/2.f ;
                     // ok, first check if they are vaguely the same facings or not
                     float dz = pObjA->rb->O.z()-pObjB->rb->O.z();
                     float piover2s = dz/PiOver2;
                     float remains = piover2s - ((int)piover2s);
                     if (fabs(remains)>0.2)
                        continue;
                     dz-=remains*(PiOver2);
                     
                     if (((pObjA->flags&kGameObjScored)==0)||(pObjB->flags&kGameObjScored)==0)
                     {
                        int pts=pObjA->attach_count>pObjB->attach_count?pObjA->attach_count:pObjB->attach_count;
                        points+=pts;
                        pObjA->flags|=kGameObjScored;
                        pObjB->flags|=kGameObjScored;
                     }
                     // check attachements?
                     // spring force - 1 1 2 3000 0 - a b d s rest
                     int j, found=0;
                     for (j=0; j<pObjA->attach_count; j++)
                        if (pObjA->attached_list[j]==pObjB)
                        {
                           found=1;
                           break;  // we are already attached
                        }
                     if (!found)
                     {  // hey, we need to attach to our new pal
                        if (pObjA->attach_count>=6)  // really, have this increase w/score
                        {  // delete them all, score big
                           // now... delete

                           // wait for auto-crazy-reindexing of peril to end, first
                           // del_ObjGroup(pObjA);
                        }
                        else
                        {  // attach more
                           Vec2 handle_pt_a = transpose_times(pObjA->rb->R, cit->m_x - pObjA->rb->x);
                           Vec2 handle_pt_b = transpose_times(pObjB->rb->R, cit->m_x - pObjB->rb->x);
                           pObjA->cnt[pObjA->attach_count]=
                              new Pt2Pt_Constraint(Handle(*(pObjA->rb),handle_pt_a),
                                                   Handle(*(pObjB->rb),handle_pt_b));

                           sim->add(*(pObjA->cnt[pObjA->attach_count]));

                           pObjA->ang[pObjA->attach_count]=
                              new Angle_Spring_Force(Handle(*(pObjA->rb),handle_pt_a),
                                                     Handle(*(pObjB->rb),handle_pt_b),
                                                     5000.0f,100.0f, //3000.0f,2.0f,
                                                     -dz);        

                           sim->add(*(pObjA->ang[pObjA->attach_count]));

                           pObjA->attached_list[pObjA->attach_count++]=pObjB;
                           pObjB->attached_list[pObjB->attach_count++]=pObjA;
                        }
                     }
                  }
         }
      }
   }

	AJB_CONSOLE("Score " << points << endl);

	sim->draw(GL::GL_Physics2d_Draw(screen_width, camera_pos));	//GL_Physics2d_Draw is a callback object
}


void game_end_frame()
{
	PROFILER(end_frame);
	//console->draw();
	MOUSE.tick();
	KEYBOARD.tick();

	glutSwapBuffers();
	glutPostRedisplay();
	if (kErrorCheckGL) GL::gl_error_check();
}

// cybelles 482 1716

void game_draw()
{
	game_begin_frame();
	//game_tick();

	Idle() ;
	Render() ;
	
	game_end_frame();
}


//------------------------------------------------------------------------------------------------------------
// you probably don't need to edit anything below here

static bool visible = true;

static void display_CB() { if (visible) game_draw(); }

namespace Graphics {	//these needs to be exported for a few libraries
	int get_window_width() { return G_win_width; }
	int get_window_height() { return G_win_height; }
	float get_window_aspect_wh() { return float(G_win_width)/float(G_win_height); }
}
static void reshape_CB(GLsizei w, GLsizei h) 
{ 
	G_win_width = w;
	G_win_height = h;
	if (kErrorCheckGL) GL::gl_error_check();
	glViewport(0, 0, (GLsizei) w, (GLsizei) h);
	if (kErrorCheckGL) GL::gl_error_check();
}
static void visible_CB(int state) { visible = (state == GLUT_VISIBLE); }

static void mouse_CB(int button, int state, int x, int y) 
{ 
	int b=button, s=state ;

	//theGLUTApp -> GLUTModifiers = glutGetModifiers() ;

	//fSec now = GetSeconds() ;

	theGLUTApp . mousex = x ;
	theGLUTApp . mousey = y ;

	if ( b == GLUT_LEFT_BUTTON && s == GLUT_DOWN )
	{

		/*
		theGLUTApp -> wasLastMouseDownADoubleClick =
			( ( //!theGLUTApp->wasLastMouseDownADoubleClick )   &&
				x == theGLUTApp -> lastMouseDownX &&
				y == theGLUTApp -> lastMouseDownY //&&
				//now - theGLUTApp -> GetMouseDownWhen() < theGLUTApp -> GetDoubleClickDelta()
				) ;*/

		theGLUTApp . lastMouseDownX = x ;
		theGLUTApp . lastMouseDownY = y ;

		MouseDown( GetMouseLoc() ) ;

		theGLUTApp.isMouseDown = true ;
	}

	if ( b == GLUT_LEFT_BUTTON && s == GLUT_UP ) {
	
		theGLUTApp . wasClickInPlace =
					( x == theGLUTApp.lastMouseDownX &&
					  y == theGLUTApp.lastMouseDownY ) ;

		MouseUp( GetMouseLoc() ) ;

		theGLUTApp.isMouseDown = false ;
	}

	//
	if (button == GLUT_LEFT_BUTTON)
		MOUSE.set_state_left(state != GLUT_UP);
	else if (button == GLUT_RIGHT_BUTTON)
		MOUSE.set_state_right(state != GLUT_UP);
	MOUSE.set_pos(Vec2(x, y));
}

static void mouseMoved_CB        (int x, int y) 
{
	MOUSE.set_pos(Vec2(x,y));
	
	MouseMoved( fPoint2( x-theGLUTApp.mousex, y-theGLUTApp.mousey ) ) ;

	theGLUTApp.mousex=x;
	theGLUTApp.mousey=y;
}

//static void motion_CB        (int x, int y) { MOUSE.set_pos(Vec2(x,y)); theGLUTApp.mousex=x; MouseMoved( fPoint2( x-theGLUTApp.mousex, y-theGLUTApp.mousey ) ) ; theGLUTApp.mousey=y; }
//static void passive_motion_CB(int x, int y) { MOUSE.set_pos(Vec2(x,y)); theGLUTApp.mousex=x; MouseMoved( fPoint2( x-theGLUTApp.mousex, y-theGLUTApp.mousey ) ) ; theGLUTApp.mousey=y; }

static void keyboard_CB   (unsigned char c, int x, int y) {
	if (c == '`')
		launchEditor("bodies.physdb");
    KEYBOARD.set_down(Keyboard::upcase(c));
}
static void keyboard_up_CB(unsigned char c, int x, int y) { KEYBOARD.set_up  (Keyboard::upcase(c)); }

static void special_CB   (int c, int x, int y) { if (c < 256) KEYBOARD.set_down(c | 0x80); }
static void special_up_CB(int c, int x, int y) { if (c < 256) KEYBOARD.set_up  (c | 0x80); }

int glut_main(int argc, char** argv) 
{
	glutInit(&argc, argv);

	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
	glutCreateWindow("IGJ2 GLUT Sample");
	glutReshapeWindow( kPixelWidth, kPixelHeight );

	glutIgnoreKeyRepeat(true);
	glutDisplayFunc(display_CB);
	glutReshapeFunc(reshape_CB);
	glutVisibilityFunc(visible_CB);
	glutKeyboardFunc(keyboard_CB);
	glutKeyboardUpFunc(keyboard_up_CB);
	glutMouseFunc(mouse_CB);
	glutMotionFunc(mouseMoved_CB);
	glutPassiveMotionFunc(mouseMoved_CB);
	glutSpecialFunc(special_CB);
	glutSpecialUpFunc(special_up_CB);

	glClearColor( 1, 1, 1, 0 ) ;
	glEnable( GL_LINE_SMOOTH ) ;
	glEnable( GL_BLEND ) ;
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA ) ;

	if (kErrorCheckGL) GL::gl_error_check();	
	delete new int; delete[] new int[1]; AJB_DELETE AJB_NEW int; AJB_DELETE[] AJB_NEW int[1]; //to init profilers in a nice order
	PROFILER_ROOT(root);
	JOYSTICK.init();
	console = new GL::Console_GL;
	game_init();
	if (kErrorCheckGL) GL::gl_error_check();

	glutMainLoop();
	return 0;
}
}

int main(int argc, char** argv) 
{
	return ajb::glut_main(argc, argv);
}
