// 2004 - Chaim Gingold

#include "PlantSeg.h"
#include "fColor.h"
#include <assert.h>
#include "chaim.h"

#include "Plant.h"

PlantWorld::PlantWorld()
{
	sim = 0 ;
	drag = 0 ;
	gravity = 0 ;
	earth = 0 ;

	for( int i=0; i<kMaxObjs; i++ )
	{
		objs[i] = 0 ;
	}
}

void PlantWorld::DropObjects ( PlantObj* adjTo )
{
	fPoint2 p = adjTo->GetLoc() ;
	float r = adjTo->GetRadius() ;

	for( int i=0; i<kMaxObjs; i++ )
	{
		if ( objs[i] == adjTo ) continue ;
		if ( !objs[i] ) continue ;

		if ( objs[i]->GetLoc().Dist(p) < pix(1) + r + objs[i]->GetRadius() )
		{
			objs[i]->letGo = true ;
			//objs[i]->MoveTo( 
		}
	}
}

void PlantWorld::Init ( void )
{
	// make simulator
	sim = new Sim ;
	if (!sim) return ;

	// a little drag helps with stability
	drag = new Drag_Force( kDragForce, kDragForce ) ; 

	if (drag) sim->add(*drag) ;


	// make gravity
	float g = kGravity ;

	gravity = new Gravity_Force(Vec2(0.0f, g)) ;

	if (gravity) sim->add(*gravity) ;

	// make da world
    earth = new Rigid_Body( make_Body_Null(), Vec2::zero(),0.0f,0.0f,0.0f ) ;
	sim->add(*earth) ;

	// make the ground
	fRect gb = GetBounds() ;

	fRect gbl=gb, gbr=gb, gbt=gb ;

	gb.top = gb.bottom - kGroundHeight ;

	gbl.right = gbl.left  + kWallWidth ;
	gbr.left  = gbr.right - kWallWidth ;
	gbt.bottom= gbt.top   + kWallWidth ;

	groundBounds = gb ;

	ground  = NewRigidBox( gb  ) ;
	groundl = NewRigidBox( gbl ) ;
	groundr = NewRigidBox( gbr ) ;
	groundt = NewRigidBox( gbt ) ;

	if ( ground )
	{
		ground -> inv_m = 0 ;
		ground -> inv_I = 0 ;

		groundl -> inv_m = 0 ;
		groundl -> inv_I = 0 ;

		groundr -> inv_m = 0 ;
		groundr -> inv_I = 0 ;

		groundt -> inv_m = 0 ;
		groundt -> inv_I = 0 ;

		sim->add(*ground) ;
		sim->add(*groundl) ;
		sim->add(*groundr) ;
		sim->add(*groundt) ;
	}

	// make plant
	plant = new Plant( this, fPoint2( GetBounds().GetCenterX(), GetBounds().bottom-kGroundHeight ) ) ;

	// make some fruit
	//plant->GrowFruitAtAllTips() ;

	fRect fruitBox = GetBounds() ;
	//fruitBox.Shift( 0, -fruitBox.GetHeight() ) ;
	//fruitBox.bottom = gb.top ;
	fruitBox.Inset( kGroundHeight*3, kGroundHeight ) ;

	for( int i=0; i<10; i++ )
	{
		fPoint2 q = fruitBox.GetRandomPoint() ;

		float cx = fruitBox.GetCenterX() ;

		float l = flerp( fruitBox.left, fruitBox.right, .25 ) ;
		float r = flerp( fruitBox.left, fruitBox.right, .75 ) ;

		if ( q.x >= l && q.x <= r )
		{
			if ( q.x < cx ) q.x = l ;
			else q.x = r ;
		}

		q.x += GetRandom(pix(3)) ;

		AddObj( new Fruit( this, q ) ) ;
	}
}

PlantWorld::~PlantWorld()
{
	for( int i=0; i<kMaxObjs; i++ )
	{
		if ( objs[i] ) delete objs[i] ;
	}	

	if (ground)
	{
		sim->remove(*ground) ;
		delete ground ;
	}

	if (groundl)
	{
		sim->remove(*groundl) ;
		delete groundl ;
	}

	if (groundr)
	{
		sim->remove(*groundr) ;
		delete groundr ;
	}

	if (groundt)
	{
		sim->remove(*groundt) ;
		delete groundt ;
	}

	delete plant ;

	delete sim ;
}

void
PlantWorld::EraseObjs ( void )
{
	for( int i=0; i<kMaxObjs; i++ )
	{
		if ( objs[i] )
		{
			delete objs[i] ;
			objs[i] = 0 ;
		}
	}
}

void
PlantWorld::AddObj ( PlantObj* o )
{
	int i = FindObj(0) ;
	if ( i!=-1 ) objs[i] = o ;
}

Ant*
PlantWorld::AddAnts ( int n, fRect where )
{
	Ant * a = 0 ;

	while ( n-->0 )
	{
		int i = FindObj(0) ;

		if ( i==-1 ) return a ;

		objs[i] = a = new Ant( this, where.GetRandomPoint() ) ;
	}

	return a ;
}

PlantObj*
PlantWorld::PickObj ( fPoint2 p ) const
{
	for( int i=0; i<kMaxObjs; i++ )
	{
		if ( objs[i] && objs[i]->Pick(p) ) return objs[i] ;
	}

	return 0 ;
}

int
PlantWorld::FindObj ( PlantObj* q ) const
{
	for( int i=0; i<kMaxObjs; i++ )
	{
		if ( objs[i] == q ) return i ;
	}

	return -1 ;
}



void PlantWorld::RenderObjs ( void )
{
	if ( plant )
	{
		plant -> Render() ;
	}
	

	//PenWidth( pix(1) ) ;

	fColor::brown()->SetColor() ;
	FrameRigidBody( ground  ) ;
	//FrameRigidBody( groundl ) ;
	//FrameRigidBody( groundr ) ;
	//FrameRigidBody( groundt ) ;

	fColor::black()->SetColor() ;
	Line( groundBounds.GetTopLeft(), groundBounds.GetTopRight() ) ;


	for( int i=0; i<kMaxObjs; i++ )
	{
		if ( objs[i] ) objs[i] -> Render() ;
	}
}

void PlantWorld::UpdateObjs ( void )
{
	for( int i=0; i<kMaxObjs; i++ )
	{
		if ( objs[i] )
		{
			objs[i] -> Update() ;

			if ( objs[i]->deleteMe )
			{
				delete objs[i] ;
				objs[i] = 0 ;
			}
		}
	}
}

fPoint2
PlantWorld::GetWaypoint ( fPoint2 _from, fPoint2 _to, float edgeClingDist ) const
{
	fPoint2 from=_from, to=_to ;

	float maxd=edgeClingDist ;
	bool fromOnPlant = plant -> PinToPlant ( &from, &maxd ) ;

	maxd=edgeClingDist ;
	bool toOnPlant = plant -> PinToPlant ( &to, &maxd ) ;

	// both not on plant
	if ( !fromOnPlant && !toOnPlant ) return to ; // -- just go there

	// go to plant
	if ( !fromOnPlant && toOnPlant ) return plant -> GetRootLoc() ;

	// leave plant -- keep going, but reset destination to root
	if ( fromOnPlant && !toOnPlant )
	{
		// we may never be able to leave the plant

		return _to ;

		fPoint2 root = plant -> GetRootLoc() ;
	
		if ( root.Dist(from) < edgeClingDist ) return _to ;

		//maxd=edgeClingDist ;
		//toOnPlant = plant -> PinToPlant ( &to, &maxd ) ;
		//assert( toOnPlant ) ;

		return root ;
	}

	// at this point, both from and to
	// are on the plant -- find out their indexes, and route

	int fromNode, toNode ;

	fromNode = plant -> GetClosestNode ( from ) ;
	toNode   = plant -> GetClosestNode ( to   ) ;

	if ( fromNode==toNode ) return to ; // just go to your destination location

	if ( fromNode==-1 || toNode==-1 ) { assert(0); return to ; } // wtf

	// find waypoint on plant
	toNode = plant -> FindWaypointNode( fromNode, toNode ) ;

	if ( toNode==-1 ) return to ;
	else return plant -> GetNodeLoc( toNode ) ;
}

bool PlantWorld::PinToEdge ( fPoint2 * p , float maxd,
							 Rigid_Body** _rb, bool * onGround,
							 bool dontLetPlantGo,
							 bool onlyAdjEdges,
							 int oldPlantEdge[2], float oldPlantFudge
							 )
{
	if (onlyAdjEdges)
	{
		int x=1 ;
	}

	if (_rb) *_rb = 0 ;

	if (onGround) *onGround=false ;

	//
	int res=0 ; // 0: nothing, 1: plant, 2: ground

	fPoint2 plantPin = *p ;


	if ( plant->PinToPlant ( &plantPin, &maxd, onlyAdjEdges, oldPlantEdge, oldPlantFudge ) )
	{
		res=1 ;
	}

	fPoint2 groundPin = *p ;

	if ( !dontLetPlantGo )
	{
		if ( PinToGround( &groundPin, &maxd ) )
		{
			if (onGround) *onGround = true ;

			res=2 ;
		}
	}

	// get rigid body to pin to for plant
	if ( res==1 )
	{
		float bd ;

		if (_rb) *_rb = plant->FindClosestPlantBody( *p, &bd, onlyAdjEdges ? oldPlantEdge : 0 ) ;
	}
	else if ( oldPlantEdge )
	{
		oldPlantEdge[0]=-1 ;
		oldPlantEdge[1]=-1 ;
	}

	// return
	if ( res==1 ) *p = plantPin ;
	else if ( res==2 ) *p = groundPin ;

	return res>0 ;
}

bool PlantWorld::PinToGround ( fPoint2 * p , float * maxd )
{
	fPoint2 a, b ;

	a = groundBounds.GetTopLeft()  ;
	b = groundBounds.GetTopRight() ;

	a.y -= kAntRadius ;
	b.y -= kAntRadius ;

	float d = p->DistToLineSegment( a, b ) ;

	if ( d < *maxd )
	{
		*maxd = d ;
		p->SnapToLineSegment( a, b ) ;
		return true ;
	}

	return false ;
}

Rigid_Body* PlantWorld::FindClosestRigidBodyToStandOn ( fPoint2 loc )
{
	float d ;
	return plant->FindClosestPlantBody( loc, &d, 0 ) ;
}

void PlantWorld::CommandSelected ( int p, Handle goHere )
{
	for( int i=0; i<kMaxObjs; i++ )
	{
		if ( objs[i] && objs[i]->selected == p )
		{
			objs[i]->WalkToHere( goHere ) ;
		}
	}
}

void PlantWorld::UnselectAll ( int p )
{
	for( int i=0; i<kMaxObjs; i++ )
	{
		if ( objs[i] && objs[i]->selected == p ) objs[i]->selected = 0 ;
	}
}

void PlantWorld::Select	( int p, fPoint2 loc )
{
	for( int i=0; i<kMaxObjs; i++ )
	{
		if ( objs[i] && objs[i]->Pick(loc) ) objs[i]->selected = p ;
	}
}

void PlantWorld::Select	( int p, fRect r )
{
	for( int i=0; i<kMaxObjs; i++ )
	{
		if ( objs[i] )
		{
			fRect q = r ;
			q.Inset( -objs[i]->GetRadius(), -objs[i]->GetRadius() ) ;

			if ( r.DoesContain( objs[i]->GetLoc() ) ) objs[i]->selected=p ;
		}
	}
}



	/*
	 *	PlantObj
	 */

PlantObj::PlantObj( PlantWorld * world, fPoint2 loc ) :
	onBody(0) ,
	world(world) ,
	_loc(loc)
{
	jumpPlantEdges=false ;
	deleteMe=false ;

	assert( world ) ;
	assert( world->sim ) ;

	selected = false ;
	
	letGo = false ;

	oldPlantEdges[0] = -1 ;
	oldPlantEdges[1] = -1 ;
}

PlantObj::~PlantObj()
{
	if (me)
	{
		world->sim->remove(*me) ;
		delete me ;
	}

	if (pt2pt)
	{
		world->sim->remove(*pt2pt) ;
		delete pt2pt ;
	}
}

void
PlantObj::Init ( PlantWorld * world, fPoint2 loc )
{
	world->PinToEdge( &loc, 10000 ) ;
	this->_loc = loc ;


	me = NewRigidCircle( loc, GetRadius(), kAntMass ) ;

	if (me)
	{
		me -> inv_I = 0 ; // don't roll
		me -> m_Pbody -> m_collision_plane = kObjCollisionMask ;
		world->sim->add( *me ) ;
	}

	onBody = 0 ;
	pt2pt  = 0 ;

	MoveTo( loc, true, false ) ;
}

bool PlantObj::MoveTo ( fPoint2 dst, bool pinToEdge, bool dontLetPlantGo )
{
	bool res ;
	
	onGround = false ;

	// find what edge to stick to
	if (pinToEdge)
	{
		if ( jumpPlantEdges )
		{
			res = world->PinToEdge( &dst, GetRadius()*3, &onBody, &onGround, dontLetPlantGo, false, oldPlantEdges ) ;
		}
		else
		{		
			res = world->PinToEdge( &dst, GetRadius()*2, &onBody, &onGround, dontLetPlantGo, true, oldPlantEdges ) ;
		}
	}
	else
	{
		onBody=0 ;
		res = false ;
	}

	// setup physics constraints for motion
	bool rmvPt2Pt = false ;

	if ( res )
	{
		_loc = dst ;

		if ( onBody )
		{
			Handle meHandle( *me, Vec2(0,0) ) ;
			Handle itHandle( *onBody, f2v( ws2bs( onBody, _loc ) ) ) ;//GetLoc() ) ) ) ;

			if ( !pt2pt )
			{
				pt2pt = new Pt2Pt_Constraint( meHandle, itHandle ) ;

				if (pt2pt) world->sim->add( *pt2pt ) ;
			}
			else 
			{
				pt2pt -> m_b = itHandle ;
			}
		} else rmvPt2Pt = true ;
	}
	else rmvPt2Pt = true ;

	if ( rmvPt2Pt )
	{
		// teleport
		me->x = f2v(dst) ;
		me->update_derived_state() ;
	}

	if ( rmvPt2Pt && pt2pt )
	{
		world->sim->remove( *pt2pt ) ;
		delete pt2pt ;
		pt2pt = 0 ;
	}

	return res ;
}

void PlantObj::Update ( void )
{
	bool on = MoveTo( GetLoc(), !letGo, !letGo && onBody ) ;
	//bool on = MoveTo( GetLoc(), true, !letGo && onBody ) ;
	
}

void PlantObj::Render ( void )
{
	GetColor().SetColor() ;
	PaintCircle( GetLoc(), GetRadius() ) ;

	if ( selected )
	{
		if ( selected==1 ) fColor::yellow() -> SetColor() ;
		else fColor::blue() -> SetColor() ;

		FrameCircle( GetLoc(), GetRadius() + pix(4) ) ;
		FrameCircle( GetLoc(), GetRadius() + pix(6) ) ;
	}

	fColor::black()->SetColor() ;
	FrameCircle( GetLoc(), GetRadius() ) ;
}

fPoint2
PlantObj::GetLoc ( void ) const
{
	return bs2ws( me, fPoint2(0,0) ) ;

	//return loc ;
	//return onBody->GetBase() ;
}

bool PlantObj::Pick ( fPoint2 p ) const
{
	return GetLoc().Dist(p) <= GetRadius() ;
}

void Ant::WalkRelative ( fPoint2 p )
{
	if ( !onBody && !onGround )
	{
		if ( p.y < 0 ) p.y=0 ;

		p.Scale(.1) ;
	}
	else if ( onGround )
	{
		if ( letGo )
		{
			p.x *= .5 ;
			p.y = -pix(64) ;
		}
		else p.x *= .2 ;
	}

	{
		p.Add( GetLoc() ) ;

		WalkToHere( Handle( *world->earth, f2v( ws2bs( world->earth, p ) ) ) ) ;
	}
}

/*
void Ant::Render ( void )
{
	PlantObj::Render() ;

	if (0) // walking )
	{
		fColor::gray()->SetColor() ;

		fPoint2 me = GetLoc() ;

		fPoint2 waypoint = world->GetWaypoint( me, GetfDst(), GetRadius() ) ;
		Line( me, waypoint ) ;

		fColor::brown()->SetColor() ;
		Line( waypoint, GetfDst() ) ;
	}
}*/

fPoint2
Ant::GetfDst ( void )
{
	if ( walking ) return v2f( dst.x_worldS() ) ;
	else return GetLoc() ;
}

void Ant::Update ( void )
{
	if (letGo) letGo-- ;

	if (push) world->DropObjects ( this ) ;

	PlantObj::Update() ;


	fPoint2 dstLoc = GetfDst() ;

	fPoint2 waypoint = world->GetWaypoint( GetLoc(), dstLoc, GetRadius()*.25 ) ;

	Handle dstH( *world->earth, f2v( ws2bs( world->earth, waypoint ) ) ) ;


	if ( walking && GetLoc().Dist( dstLoc ) < GetRadius() ) walking = false ;


	if ( walking && !moveSpring )
	{
		moveSpring = new Spring_Force(
			Handle( *me, Vec2(0,0) ) ,
			dstH,
			256.0f,		//Kspring
			1.0f,		//Kdamp
			0.0f		//rest length
		) ;

		world->sim->add(*moveSpring) ;
	}
	else if ( !walking && moveSpring )
	{
		world->sim->remove(*moveSpring) ;
		delete moveSpring ;
		moveSpring=0;
	}
	else if ( walking && moveSpring )
	{
		moveSpring->m_b = dstH ;
	}


	/*
	if ( walking )
	{
		fPoint2 me = GetLoc() ;

		if ( me.Dist(dst) < GetRadius() ) walking = false ;
		else
		{
			// walk towards waypoint
			fPoint2 waypoint = world->GetWaypoint( me, dst, GetRadius() ) ;

			fPoint2 stepTo = me ;
			
			stepTo.Add( me.Angle(waypoint), kAntStep ) ;

			MoveTo( stepTo, false ) ;
		}
	}*/
}

Fruit::Fruit( PlantWorld* w, fPoint2 loc ) : PlantObj(w,loc)
{
	int x=1 ;

	Init(w,loc) ;

	droppedByAnt=0 ;

	jumpPlantEdges=true ;
}

/*
void Fruit::Render ( void )
{

}
*/
void Fruit::Update ( void )
{
	letGo = world->plant->fruitDeadZone.DoesContain( GetLoc() ) ;

	PlantObj::Update() ;

	if ( world->plant->FitFruit( GetLoc() ) )
	{
		deleteMe = true ;
		
		if ( world->plant->CountFruitSlots()==0 ) music.overlay( "LevelUp", 0 ) ;
		else music.overlay( "Target", 0 ) ;

	}
}