#ifndef __PlantSegH__
#define __PlantSegH__
// 2004 - Chaim Gingold

#include "SimHelp.h"
#include "chaim.h"
#include "fColor.h"

class PlantWorld ;
class PlantSeg ;
class PlantObj ;
class Ant ;

const float kAntStep	 = pix(2) ;

const float kAntMass	 = .1 ;
const float kAntRadius   = pix(8) ;
const int   kNumAnts     = 1 ;

const float kFruitRadius = pix(12) ;

const int   kPlantGrowDepth = 8 ;

const float kSimTimeStep  = 1.0f/30.0f ;
const float kGroundHeight = pix(32) ;
const float kWallWidth	  = pix(8)  ;

const float kPlantObjDefForce	= 400.0f ;
const float kGravity			= 20.0f ;
const float kDragForce			= .05f   ;

const float kPlantDensity		= 1 ;

/*
const float kPlantSpringK		= 10000 ;
const float kPlantSpringDamp	= 1 ;
const float kPlantSpringRestLen = .5 ;

const float kPlantStandSpringK		 = 5 ;
const float kPlantStandSpringDamp	 = 1 ;
const float kPlantStandSpringRestLen = 100 ;
*/


	/*
	 *	PlantWorld
	 */

class Plant ;

class PlantWorld
{

public:

	PlantWorld() ;
	~PlantWorld() ;

	void Init ( void ) ;
	void RenderObjs ( void ) ;
	void UpdateObjs ( void ) ;

	fPoint2 GetWaypoint ( fPoint2 from, fPoint2 to, float edgeClingDist ) const ;

	Sim				* sim ;
	Drag_Force		* drag ;
	Gravity_Force	* gravity ;

	Rigid_Body		* earth ; // large invisible non-moving mass

	void EraseObjs ( void ) ;

	void AddObj	   ( PlantObj* ) ;
	Ant* AddAnts   ( int n, fRect where ) ;

	PlantObj* PickObj ( fPoint2 ) const ;

	bool PinToEdge ( fPoint2 * p , float maxd,
			Rigid_Body** standOnBody=0,
			bool * onGround=0, bool dontLetPlantGo=false,
			bool onlyAdjEdges=false,
			int oldPlantEdge[2]=0, float oldPlantFudge=kFruitRadius/2 //*2
			) ;

	// writes new edges in to oldPlantEdge

	Rigid_Body* FindClosestRigidBodyToStandOn ( fPoint2 loc ) ;


	Plant * plant ;


	void CommandSelected ( int, Handle goHere ) ;

	void UnselectAll ( int ) ;
	void Select		 ( int, fPoint2 p ) ;
	void Select		 ( int, fRect ) ;

	void DropObjects ( PlantObj* adjTo ) ; // but don't drop adjTo

	fRect groundBounds ;

private:
	bool PinToGround ( fPoint2 * p , float * maxd ) ;

	int FindObj ( PlantObj* ) const ;

	enum
	{
		kMaxObjs = 128
	} ;
	PlantObj * objs[ kMaxObjs ] ;

	Rigid_Body * ground , *groundl, *groundr, *groundt ;

} ;



	/*
	 *	PlantObj
	 */

class PlantObj
{

public:
	PlantObj( PlantWorld*, fPoint2 loc ) ;
	virtual ~PlantObj() ;

	bool    Pick   ( fPoint2 ) const ;
	fPoint2 GetLoc ( void ) const ;

	virtual float GetRadius ( void ) const { return 18 ; }
	virtual fColor GetColor ( void ) const { return *fColor::green() ; }

	virtual float GetGravitationalForce ( void ) const { return kPlantObjDefForce ; }

	bool MoveTo ( fPoint2, bool pinToEdge, bool dontLeavePlant ) ; // whether succesful or not -- will fall if not!

	virtual void Render ( void ) ;
	virtual void Update ( void ) ;

	int selected ; // 0: none, 1: player 1, 2: player 2

	virtual void WalkToHere ( Handle ) {}

	bool deleteMe ;

	bool jumpPlantEdges ;

	int letGo ;
	bool onGround ;

	Rigid_Body	     * onBody ;

	int player ;

	int oldPlantEdges[2] ;

protected:
	void Init ( PlantWorld * world, fPoint2 loc ) ;

	PlantWorld	   * world ;

	Rigid_Body	     * me ;

private:

	fPoint2			  _loc ;

	Pt2Pt_Constraint * pt2pt ;

} ;

class Fruit : public PlantObj
{
public:
	Fruit( PlantWorld* w, fPoint2 loc ) ;

	float GetRadius ( void ) const { return kFruitRadius ; }
	virtual fColor GetColor ( void ) const { return *fColor::orange() ; }

	//virtual void Render ( void ) ;
	virtual void Update ( void ) ;

	Ant * droppedByAnt ;

} ;

class Ant : public PlantObj
{
public:
	Ant( PlantWorld* w, fPoint2 loc ) : PlantObj(w,loc)
	{
		Init(w,loc) ;
		walking = false ;
		moveSpring = 0 ;
		push = false ;
	}

	virtual void WalkToHere ( Handle h ) { dst=h ; walking=true; }

	virtual void WalkRelative ( fPoint2 ) ;

	float GetRadius ( void ) const { return kAntRadius ; }
	virtual fColor GetColor ( void ) const { return player%2 ? *fColor::blue() : *fColor::green() ; }

	//virtual void Render ( void ) ;
	virtual void Update ( void ) ;

	bool push ;

private:
	fPoint2 GetfDst ( void ) ;

	Spring_Force	 * moveSpring ;

	bool walking ;

	//fPoint2 dst ;
	Handle dst ;

} ;



#endif